/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Stack;

import org.eclipse.emf.common.command.BasicCommandStack;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.AdapterImpl;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.emfstore.client.model.Configuration;
import org.eclipse.emf.emfstore.client.model.ModelPackage;
import org.eclipse.emf.emfstore.client.model.ProjectSpace;
import org.eclipse.emf.emfstore.client.model.Workspace;
import org.eclipse.emf.emfstore.client.model.WorkspaceManager;
import org.eclipse.emf.emfstore.client.model.impl.ProjectSpaceImpl;
import org.eclipse.emf.emfstore.client.model.observers.OperationListener;
import org.eclipse.emf.emfstore.client.model.util.EMFStoreCommand;
import org.eclipse.emf.emfstore.server.model.versioning.operations.AbstractOperation;
import org.fortiss.tooling.kernel.services.ICommandStackService;

/**
 * This class implements the {@link ICommandStackService} interface.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @levd.rating RED Rev:
 */
public class CommandStackService implements ICommandStackService {

	/** Stores the EMFStore command stack. */
	private BasicCommandStack emfStoreCommandStack;

	/** Stores the project state observers. */
	private Map<ProjectSpace, ProjectSpaceObserver> projectSpaceObserverMap = new HashMap<ProjectSpace, CommandStackService.ProjectSpaceObserver>();

	/** Stores the workspace listener. */
	private AdapterImpl workspaceListener;

	/** Constructor. */
	public CommandStackService() {

		Workspace ws = WorkspaceManager.getInstance().getCurrentWorkspace();
		createWorkspaceListener();
		ws.eAdapters().add(workspaceListener);

		for (ProjectSpace ps : ws.getProjectSpaces()) {
			ProjectSpaceObserver state = new ProjectSpaceObserver(ps);
			projectSpaceObserverMap.put(ps, state);
		}

		emfStoreCommandStack = (BasicCommandStack) Configuration
				.getEditingDomain().getCommandStack();
	}

	/** {@inheritDoc} */
	@Override
	public void runAsCommand(EObject target, final Runnable runner) {
		EMFStoreCommand command = new EMFStoreCommand() {

			@Override
			protected void doRun() {
				runner.run();
			}
		};
		emfStoreCommandStack.execute(command);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canUndo(EObject target) {
		ProjectSpace projectSpace = WorkspaceManager.getProjectSpace(target);
		return !projectSpace.getOperations().isEmpty();
	}

	/** {@inheritDoc} */
	@Override
	public boolean canRedo(EObject target) {
		ProjectSpace projectSpace = WorkspaceManager.getProjectSpace(target);
		return projectSpaceObserverMap.get(projectSpace).canRedo();
	}

	/** {@inheritDoc} */
	@Override
	public void undo(EObject target) {
		final ProjectSpace projectSpace = WorkspaceManager
				.getProjectSpace(target);
		// also pushes the operation on the project state observer's redo stack
		emfStoreCommandStack.execute(new EMFStoreCommand() {

			@Override
			protected void doRun() {
				projectSpace.undoLastOperation();
			}
		});
	}

	/** {@inheritDoc} */
	@Override
	public void redo(EObject target) {
		ProjectSpace projectSpace = WorkspaceManager.getProjectSpace(target);
		projectSpaceObserverMap.get(projectSpace).redo();
	}

	/** Creates the workspace listener. */
	private void createWorkspaceListener() {
		workspaceListener = new AdapterImpl() {
			@Override
			public void notifyChanged(Notification msg) {
				if ((msg.getFeatureID(Workspace.class)) == ModelPackage.WORKSPACE__PROJECT_SPACES) {
					if (msg.getEventType() == Notification.ADD
							&& ModelPackage.eINSTANCE.getProjectSpace()
									.isInstance(msg.getNewValue())) {
						ProjectSpace projectSpace = (ProjectSpace) msg
								.getNewValue();
						ProjectSpaceObserver obs = new ProjectSpaceObserver(
								projectSpace);
						projectSpaceObserverMap.put(projectSpace, obs);
					} else if (msg.getEventType() == Notification.REMOVE
							&& ModelPackage.eINSTANCE.getProjectSpace()
									.isInstance(msg.getOldValue())) {
						ProjectSpace projectSpace = (ProjectSpace) msg
								.getOldValue();
						ProjectSpaceObserver obs = projectSpaceObserverMap
								.remove(projectSpace);
						obs.dispose();
					}
				}
			}
		};
	}

	/** Observes the state of a {@link ProjectSpace}. */
	private class ProjectSpaceObserver implements OperationListener {

		/** Stores the project space reference. */
		private final ProjectSpace projectSpace;

		/** Stores the redo stack. */
		private final Stack<AbstractOperation> redoStack = new Stack<AbstractOperation>();

		/** Flag indicating the current operation is a redo. */
		private boolean redoInProgress = false;

		/** Constructor. */
		public ProjectSpaceObserver(ProjectSpace projectSpace) {
			this.projectSpace = projectSpace;
			projectSpace.addOperationListener(this);
		}

		/** {@inheritDoc} */
		@Override
		public void operationExecuted(AbstractOperation operation) {
			if (!redoInProgress) {
				// clear stack if not observing a redo
				redoStack.clear();
			}
		}

		/** {@inheritDoc} */
		@Override
		public void operationUnDone(AbstractOperation operation) {
			redoStack.add(operation);
		}

		/** Returns whether redo is possible. */
		public boolean canRedo() {
			return !redoStack.isEmpty();
		}

		/** Redo the last operation. */
		public void redo() {
			final LinkedList<AbstractOperation> ops = new LinkedList<AbstractOperation>();
			ops.add(redoStack.pop());
			emfStoreCommandStack.execute(new EMFStoreCommand() {

				@Override
				protected void doRun() {
					redoInProgress = true;
					((ProjectSpaceImpl) projectSpace).applyOperations(ops);
					redoInProgress = false;
				}
			});
		}

		/** Disposes the project state observer. */
		public void dispose() {
			projectSpace.removeOperationListener(this);
		}
	}
}
